#include <linux/module.h>
#include <linux/types.h>
#include <linux/fs.h>
#include <linux/init.h>
#include <linux/vfs.h>
#include <linux/errno.h>
#include <linux/pagemap.h>
#include <linux/smp_lock.h>
#include <linux/buffer_head.h>
#include <linux/slab.h>
#include "akai.h"

// Pre: d != NULL
// Read the partitions and store the offsets relative to the filesystem
// in a struct.
// returns true if we have at least one valid partition, false otherwise
int init_ptable(struct super_block* sb, struct akaifs_layout* d)
{
  unsigned int curblock = 0;
  int valid = 1;
  int i = 0;
  struct buffer_head* bh;
  struct s1000_partheader* par;

  // Reset number of partitions
  d->number = 0;

  do
    {
      bh = sb_bread(sb, REALBLOCK(curblock));

      valid = ((bh) ? 1 : 0);

      if(valid)
	{
	  par = (struct s1000_partheader*) bh->b_data;
	  valid = ((le16_to_cpu(par->size) <= MAXPARSIZE_SECTORS) ? 1 : 0);
	  if(valid)
	    {
	      d->offsets[i] = curblock;
	      d->number++;
	      // Next partition
	      curblock += le16_to_cpu(par->size);
	    }
	  brelse(bh);
	}
      i++;
    }
  while(valid && (i < MAXNUMPARTITIONS));
  
  return (d->number != 0);
}

// Resolve a volume by name to an index in the partitionheader
// True if it resolves, false otherwise
int resolve_vol_byname(struct super_block* sb, int p, char* name, int* result)
{
  int found = 0;
  int i = 0;
  struct s1000_partheader* par = NULL;
  struct akaifs_layout* layout = (struct akaifs_layout*) sb->s_fs_info;
  struct buffer_head* bh;
  
  bh = sb_bread(sb, REALBLOCK(layout->offsets[p]));
  
  if(bh)
    {
      par = (struct s1000_partheader*) bh->b_data;

      while( (i<100) & !found )
	{
	  if(strncmp(par->volumes[i].name, name, 12) == 0)
	    {
	      found = 1;
	      *result = i;
	    }
	  i++;
	}
      brelse(bh);
    }
  return found;
}

// Resolve a volume by number: get the N-th volume in the partition
// and store filename
// returns true if there is an N-th volume, false otherwise
int resolve_vol_bynr(struct super_block* sb, int p, int nr, char* filename,
		     int* index)
{
  int i = 0;
  struct buffer_head* bh;
  struct akaifs_layout* layout = (struct akaifs_layout*) sb->s_fs_info;
  struct s1000_partheader* par;
  int found = 0;

  bh = sb_bread(sb, REALBLOCK(layout->offsets[p]));
  
  if(bh)
    {
      par = (struct s1000_partheader*) bh->b_data;

      while( (i<100) && !found )
	{
	  // type == 0 indicates an invalid volume
	  if(par->volumes[i].type != 0)
	    {
	      if(nr==0)
		{
		  if(filename)
		    memcpy(filename,par->volumes[i].name, 12);
		  if(index)
		    *index = i;
		  found = 1;
		}
	      nr--;
	    }
	  i++;
	}
      brelse(bh);
    }
  return found;
}

// Resolve a file-index by name
// Currently we assume the page to be 4K, which means that
// an entire volume-header fits into one page
int resolve_file_byname(struct super_block* sb, int p, int v,
			 char* filename, int* index)
{
  struct buffer_head* bhp;
  struct buffer_head* bhv;
  struct akaifs_layout* layout = (struct akaifs_layout*) sb->s_fs_info;
  struct s1000_partheader* par = NULL;
  struct s1000_direntry* entry;
  int found = 0;
  int i = 0;

  bhp = sb_bread(sb, REALBLOCK(layout->offsets[p]));
  if(bhp)
    {
      par  = (struct s1000_partheader*) bhp->b_data;
      bhv = sb_bread(sb,REALBLOCK(layout->offsets[p] + par->volumes[v].start));
      if(bhv)
	{
	  entry = (struct s1000_direntry*) bhv->b_data;
	  
	  while( (i < MAX_DIRENTRIES) && !found)
	    {
	      if(strncmp(entry->filename, filename, 12)==0)
		{
		  if(index)
		    *index = i;
		  found = 1;
		}
	      i++;
	      entry++;
	    }
	  brelse(bhv);
	}
      brelse(bhp);
    }
  return found;
}
// Analogous to volume_bynr, but for files
int resolve_file_bynr(struct super_block* sb, int p, int nr, int v,
		      char* name, int* index)
{
  struct buffer_head* bhp;
  struct buffer_head* bhv;
  struct akaifs_layout* layout = (struct akaifs_layout*) sb->s_fs_info;
  struct s1000_partheader* par = NULL;
  struct s1000_direntry* entry;
  int found = 0;
  int i = 0;

  bhp = sb_bread(sb, REALBLOCK(layout->offsets[p]));
  if(bhp)
    {
      par  = (struct s1000_partheader*) bhp->b_data;
      bhv = sb_bread(sb,REALBLOCK(layout->offsets[p] + par->volumes[v].start));
      if(bhv)
	{
	  entry = (struct s1000_direntry*) bhv->b_data;

	  while( (i<MAX_DIRENTRIES) && !found )
	    {
	      if(entry->filetype != 0)
		{
		  if(nr==0)
		    {
		      if(name)
			{
			  memcpy(name, entry->filename, 12);
			}
		      if(index)
			{
			  *index = i;
			}
		      found = 1;
		    }
		  nr--;
		}
	      i++;
	      entry++;
	    }
	  brelse(bhv);
	}
      brelse(bhp);
    }
  return found;
}

// Fetches information on file size and 'executable' status
void getfileprop(struct super_block* sb, int p, int v, int f,
		 unsigned long* size, int* program)
{
  struct buffer_head* bhp;
  struct buffer_head* bhv;
  struct akaifs_layout* layout = (struct akaifs_layout*) sb->s_fs_info;
  struct s1000_partheader* par = NULL;
  struct s1000_direntry* entry;

  bhp = sb_bread(sb, REALBLOCK(layout->offsets[p]));
  if(bhp)
    {
      par  = (struct s1000_partheader*) bhp->b_data;
      bhv = sb_bread(sb,REALBLOCK(layout->offsets[p] + par->volumes[v].start));
      if(bhv)
	{
	  entry = (struct s1000_direntry*) bhv->b_data;
	  if(size)
	    *size = akai_gettb(entry[f].filelength);
	  if(program)
	    *program = ((entry[f].filetype == 'p') ? 1 : 0);
	  brelse(bhv);
	}
      brelse(bhp);
    }
}

// Fetch a filesystem block in partition p, volume-index v, file-index f
// with offset, put the result in pagebuf
// Returns true if successful, false otherwise
int getfileblock(struct super_block* sb, int p, int v, int f,
	 unsigned long offset, void* pagebuf)
{
  int result = 0;
  struct buffer_head* bhp;
  struct buffer_head* bhv;
  struct buffer_head* bhf;
  struct akaifs_layout* layout = (struct akaifs_layout*) sb->s_fs_info;
  struct s1000_partheader* par = NULL;
  struct s1000_direntry* entry;
  int readblock = 0;
  int countblocks = 0;
  int blockpart = 0;
  int eof = 0;

  // 1. get the start block # of the file
  bhp = sb_bread(sb, REALBLOCK(layout->offsets[p]));
  if(bhp)
    {
      par = (struct s1000_partheader*) bhp->b_data;
      bhv = sb_bread(sb,REALBLOCK(layout->offsets[p] + par->volumes[v].start));
      brelse(bhp);
      if(bhv)
	{
	  entry = (struct s1000_direntry*) bhv->b_data;
	  readblock = le16_to_cpu(entry[f].startblock);
	  brelse(bhv);
	  // 2. calculate using the offset and USE_BLOCKSIZE
	  // which block # we should read next, including which segment
	  countblocks = offset / AKAI_BLOCKSIZE;
	  blockpart = (offset % AKAI_BLOCKSIZE) / USE_BLOCKSIZE;
	  
	  // 3. Walk through fat-table to reach that block
	  while( (countblocks!=0) && !eof )
	    {
	      readblock = fatget(sb, p, readblock);
	      if(readblock == AFAT_EOF)
		eof=1;
	      countblocks--;
	    }
	  // only if it is a valid blocknumber...
	  if(AFAT_DATA(readblock))
	    { 
	      // 4. Read it and put it in the buffer
	      bhf = sb_bread(sb,REALBLOCK(layout->offsets[p]+readblock)
			     +blockpart);
	      if(bhf)
		{
		  memcpy(pagebuf, bhf->b_data, USE_BLOCKSIZE);
		  brelse(bhf);
		  result = 1;
		}
	    }
	}
    }
  // 5. return...
  return result;
}
// Fetch the next block to read after b in partition p
// Returns the block
int fatget(struct super_block* sb, int p, int b)
{
  int res = -1;
  // Number of entries per normal sized block
  int NEPB = USE_BLOCKSIZE / sizeof(u16);
  u16* entries;
  struct buffer_head* bh;
  struct akaifs_layout* layout = (struct akaifs_layout*) sb->s_fs_info;
  // Number of entries on the partition-header page
  int nume = (USE_BLOCKSIZE - sizeof(struct s1000_partheader)) / sizeof(u16);

  if(b < nume)
    {
      bh = sb_bread(sb, REALBLOCK(layout->offsets[p]));
      if(bh)
	{
	  entries = (u16*) (bh->b_data + sizeof(struct s1000_partheader));
	  res = le16_to_cpu(entries[b]);
	  brelse(bh);
	}
    }
  else
    {
      
      b = b - nume; // offset beyond initial page

      // 1 needed because we need to locate from the next block onward...
      bh = sb_bread(sb, REALBLOCK(layout->offsets[p]) + 1
		    + (b / NEPB));

      if(bh)
	{
	  entries = (u16*) bh->b_data;
	  res = le16_to_cpu(entries[b % NEPB]);
	  brelse(bh);
	}
    }

  return res;
}
