#include <linux/module.h>
#include <linux/init.h>
#include <linux/fs.h>
#include <linux/errno.h>
#include <linux/statfs.h>
#include <linux/pagemap.h>
#include <linux/smp_lock.h>
#include "akai.h"

MODULE_LICENSE("GPL");

#define MAGIC_NUMBER 0x49bc397a /* A random number */

/* Dir operations */

static int akaifs_readdir(struct file* target, void* tofill, filldir_t filldir)
{
  struct inode* inode = target->f_dentry->d_inode;
  struct akaifs_layout*
    layout = (struct akaifs_layout*)inode->i_sb->s_fs_info;

  int res = 0;
  int offset = target->f_pos;
  char filename[13];
  int location = 0;
  filename[12] = 0;

  if( (GET_TYPE(inode->i_ino) == TYPE_ROOT) && (offset < layout->number) )
    {
      filename[0] = 'A' + offset;
      filename[1] = 0;
      filldir(tofill, filename, 1, offset,
	      CONSTRUCT(0, 0, offset, TYPE_PARTITION), DT_DIR);
      target->f_pos++;
      res = 1;
    }
  else
    if((GET_TYPE(inode->i_ino) == TYPE_PARTITION) &&
       (GET_SOURCEPARTITION(inode->i_ino) < layout->number))
      {
	if(resolve_vol_bynr(inode->i_sb, GET_SOURCEPARTITION(inode->i_ino),
			    target->f_pos, filename, &location))
	  {
	    filldir(tofill, akaii2ascii(filename,12), 12, offset,
		    CONSTRUCT(0, location, GET_SOURCEPARTITION(inode->i_ino),
			      TYPE_VOLUME), DT_DIR);
	    target->f_pos++;
	    res = 1;
	  }
      }
    else
      if((GET_TYPE(inode->i_ino) == TYPE_VOLUME) &&
	 (GET_SOURCEPARTITION(inode->i_ino) < layout->number))
	{
	  if(resolve_file_bynr(inode->i_sb, GET_SOURCEPARTITION(inode->i_ino),
			       target->f_pos,
			       GET_VOLINDEX(inode->i_ino),
			       filename, &location))
	    {
	      filldir(tofill, akaii2ascii(filename,12), 12, offset,
		      CONSTRUCT(location,
				GET_VOLINDEX(inode->i_ino),
				GET_SOURCEPARTITION(inode->i_ino),
				TYPE_FILE), DT_REG);
	      target->f_pos++;
	      res = 1;
	    }
	}

  return res;
}

static struct file_operations akaifs_diroperations =
{
  .read = generic_read_dir,
  .readdir = akaifs_readdir
};

// --

static int akaifs_readpage(struct file* file, struct page* page)
{
  struct inode* inode = page->mapping->host;
  loff_t offset;
  void* pagedata;
  int res = -EIO;

  page_cache_get(page);
  lock_kernel();
  pagedata = kmap(page);
  if (pagedata)
  {
    offset = page_offset(page);
    
    if(getfileblock(inode->i_sb,
		    GET_SOURCEPARTITION(inode->i_ino),
		    GET_VOLINDEX(inode->i_ino),
		    GET_FILEINDEX(inode->i_ino),
		    offset, pagedata))
      {
	SetPageUptodate(page);
	res = 0;
      }
    else
      {
	memset(pagedata, 0, USE_BLOCKSIZE);
	SetPageError(page);
      }
  }
  flush_dcache_page(page);
  unlock_page(page);
  kunmap(page);
  unlock_kernel();
  page_cache_release(page);

  return res;
}


/* Address operations */
static struct address_space_operations akaifs_addressoperations =
{
  .readpage = akaifs_readpage
};

// --

/* Inode operations (lookup by name) */
static struct dentry* akaifs_lookup(struct inode* inode, struct dentry* rq,
				    struct nameidata* nd)
{
  struct inode* newinode = NULL;
  struct akaifs_layout*
      layout = (struct akaifs_layout*)inode->i_sb->s_fs_info;
  int res = 0;
  char filename[12]; // not null-terminated
  int index = 0;
  
  // We're looking for a partition...
  if(GET_TYPE(inode->i_ino) == TYPE_ROOT )
    {
      if( (rq->d_name.len == 1) &&
	  ((rq->d_name.name[0] - 'A') < layout->number) )
	newinode = iget(inode->i_sb, CONSTRUCT(0,0,(rq->d_name.name[0] -'A'),
					       TYPE_PARTITION));      
    }
  else
    // Looking for a volume...
    if(GET_TYPE(inode->i_ino) == TYPE_PARTITION)
      {
	memcpy(filename, rq->d_name.name, 12);
	ascii2akaii(filename, 12);
	if(resolve_vol_byname(inode->i_sb,
			      GET_SOURCEPARTITION(inode->i_ino),
			      filename, &index))
	  {
	    newinode = iget(inode->i_sb,
			    CONSTRUCT(0,index,
				      GET_SOURCEPARTITION(inode->i_ino),
				      TYPE_VOLUME));
	  }
      }
    else
      // Looking for a file
      if(GET_TYPE(inode->i_ino) == TYPE_VOLUME)
	{
	  memcpy(filename, rq->d_name.name, 12);
	  ascii2akaii(filename, 12);
	  if(resolve_file_byname(inode->i_sb,
				 GET_SOURCEPARTITION(inode->i_ino),
				 GET_VOLINDEX(inode->i_ino),
				 filename, &index))
	    {
	      newinode = iget(inode->i_sb,
			      CONSTRUCT(index,
					GET_VOLINDEX(inode->i_ino),
					GET_SOURCEPARTITION(inode->i_ino),
					TYPE_FILE));
	    }
	}

  // Add to positive/negative cache
  d_add(rq, newinode);

  if(!newinode)
    {
      res = -ENOENT;
    }
  return ERR_PTR(res);
}

static struct inode_operations akaifs_dirinodeoperations =
{
  .lookup = akaifs_lookup
};


/* Superblock operations */

static int akaifs_statfs(struct dentry *dentry, struct kstatfs *buf)
{
  buf->f_type = MAGIC_NUMBER;
  buf->f_bsize = AKAI_BLOCKSIZE;
  buf->f_namelen = 12;

  return 0;
}

// --

static void akaifs_put_super(struct super_block* sb)
{
  /* Not much to do here
     Free the memory of the layout struct */
  kfree(sb->s_fs_info);
}

// --

void akaifs_read_inode(struct inode* inode)
{
  unsigned long fsize = 0;
  int program = 0;

  struct akaifs_layout* layout =(struct akaifs_layout*) inode->i_sb->s_fs_info;

  if(GET_TYPE(inode->i_ino) == TYPE_ROOT)
    {
      inode->i_mode = S_IFDIR+0555;
      inode->i_nlink = 1; // Should be safe
      inode->i_size = layout->number;
      inode->i_op = &akaifs_dirinodeoperations;
      inode->i_fop = &akaifs_diroperations;
    }
  else
    if((GET_TYPE(inode->i_ino) == TYPE_PARTITION) || 
       (GET_TYPE(inode->i_ino) == TYPE_VOLUME))
      {
	inode->i_mode = S_IFDIR+0555;
	inode->i_nlink = 1;
	inode->i_size = 0;
	inode->i_op = &akaifs_dirinodeoperations;
	inode->i_fop = &akaifs_diroperations;
      }
    else
      if(GET_TYPE(inode->i_ino) == TYPE_FILE)
	{
	  // find size and program status based on inode-nr
	  getfileprop(inode->i_sb,
		      GET_SOURCEPARTITION(inode->i_ino),
		      GET_VOLINDEX(inode->i_ino),
		      GET_FILEINDEX(inode->i_ino),
		      &fsize, &program);
	  
	  if(program)
	    inode->i_mode = S_IFREG+0555;
	  else
	    inode->i_mode = S_IFREG+0444;
	  inode->i_nlink = 1;
	  inode->i_size = fsize;
	  inode->i_data.a_ops = &akaifs_addressoperations;
	  inode->i_fop = &generic_ro_fops;
	}
  inode->i_mtime.tv_sec = inode->i_atime.tv_sec = inode->i_ctime.tv_sec = 0;
  inode->i_mtime.tv_nsec = inode->i_atime.tv_nsec = inode->i_ctime.tv_nsec = 0;
  inode->i_uid = inode->i_gid = 0;
}

// --

static struct super_operations akaifs_superoperations =
{
  .statfs = akaifs_statfs,
  .read_inode = akaifs_read_inode,
  .put_super = akaifs_put_super
};

// --

static int akaifs_fill_super(struct super_block* sb, void* data, int silent)
{
  int err = 0;

  // blocksize = mininum of PAGE_SIZE,AKAI_BLOCKSIZE
  sb_set_blocksize(sb, USE_BLOCKSIZE);
  sb->s_maxbytes = 0xFFFFFFFF;
  sb->s_magic = MAGIC_NUMBER;
  sb->s_flags |= MS_RDONLY;
  sb->s_op = &akaifs_superoperations;
  
  sb->s_fs_info = kmalloc(sizeof(struct akaifs_layout), GFP_KERNEL);

  if(sb->s_fs_info)
    {
      if(init_ptable(sb, (struct akaifs_layout*) sb->s_fs_info))
	{
	  sb->s_root = d_alloc_root(iget(sb,CONSTRUCT(0,0,0,TYPE_ROOT)));
	  if(!sb->s_root)
	    err = -EINVAL;
	}
      else
	err = -EINVAL;
      
      if(err)
	{
	  kfree(sb->s_fs_info);
	}
      
    }
  else
    err = -ENOMEM;
  
  return err;
}

// --

static int get_block_super(struct file_system_type* fs_type,
			   int flags, const char* dev_name,
			   void *data, struct vfsmount *mnt)
{
  return get_sb_bdev(fs_type, flags, dev_name, data, akaifs_fill_super, mnt);
}

// --

/* Module management */
static struct file_system_type akaifs_type =
{
  .owner    = THIS_MODULE,
  .name     = "akaifs",
  .get_sb   = get_block_super,
  .kill_sb  = kill_block_super,
  .fs_flags = FS_REQUIRES_DEV
};

// --

static int __init init_akaifs(void)
{
  int err = 0;
  err = register_filesystem(&akaifs_type);
  if(err)
  {
    printk(KERN_ALERT "Unable to register filesystem\n");
  }
  
  return err;
}

// --

static void __exit exit_akaifs(void)
{
  unregister_filesystem(&akaifs_type);
}

module_init(init_akaifs)
module_exit(exit_akaifs)
